"""
This file provides a JSON encoder and its corresponding decoder ('s hook).
To encode an object supported by this encoder:
json_string = json.dumps(object, cls=strenc.get_encoder())
To decode an object:
object = json.loads(json_string, object_hook=strenc.get_decoder_hook())

Note the behavior of json encode/decode will encode a subfield that is a dataclass as plain dict without
'__type__' field, as opposed to directly encoding instance of a dataclass class.
"""


import json
import dataclasses
from datetime import datetime
from typing import Any


_Encoder = None
_decoder_hook = None


def get_encoder():
    """
    Conforms to json interface json.dumps(o, cls=get_encoder())
    Returns:
        An Encoder classes
    """
    if _Encoder is None:
        init()
    return _Encoder


def get_decoder_hook():
    """
    Conforms to json interface json.loads(s, object_hook=get_decoder_hook())
    Returns:
        A decoder hook
    """
    if _decoder_hook is None:
        init()
    return _decoder_hook


def init():
    from cvpl_tools.fs import ImReadSetting, ImWriteSetting
    from cvpl_tools.array_key_dict import ArrayKeyDict
    from cvpl_tools.dataset_reference import DatapointReference, DatasetReference
    from dataclasses import fields


    # in order to support a more automatic conversion of dataclasses
    dataclasses_names = (
        ImReadSetting, ImWriteSetting, DatapointReference, DatasetReference
    )
    # here we use .__name__ to get class names without prefixes to ensure consistency of the class names
    # across Python invokations
    str_to_cls_map = {cls.__name__: cls for cls in dataclasses_names}


    def to_json_encodable(o: Any):
        if isinstance(o, ArrayKeyDict):
            d = {key: to_json_encodable(o[key]) for key in o}
            d['__type__'] = 'array_key_dict.ArrayKeyDict'
            return d
        elif isinstance(o, datetime):
            d = {
                '__type__': 'datetime.datetime',
                'v1': o.isoformat()
            }
            return d
        elif dataclasses.is_dataclass(o):
            # references:
            # https://stackoverflow.com/questions/51286748/make-the-python-json-encoder-support-pythons-new-dataclasses
            # https://stackoverflow.com/questions/53376099/python-dataclass-from-a-nested-dict/53498623#53498623
            # https://github.com/python/cpython/blob/master/Lib/dataclasses.py
            d = {'__type__': type(o).__name__}
            for f in fields(o):
                value = getattr(o, f.name)
                value = to_json_encodable(value)
                d[f.name] = value
            return d
        return o


    class EncoderDef(json.JSONEncoder):
        """
        Turns an object to string and back; support dataclasses and datetime in addition to
        common Python types.
        """
        def default(self, o: Any):
            """
            o is the object to be serialized
            """
            o = to_json_encodable(o)
            if isinstance(o, dict):  # if to_json_encodable changes o, it must end up as a dict
                return o
            return super().default(o)  # otherwise we just use default encoding


    def decoder_hook_def(o):
        if isinstance(o, dict) and '__type__' in o:
            ty = o.pop('__type__')
            if ty == 'array_key_dict.ArrayKeyDict':
                return ArrayKeyDict((key, decoder_hook_def(o[key])) for key in o)
            elif ty == 'datetime.datetime':
                return datetime.fromisoformat(o['v1'])
            else:
                cls = str_to_cls_map[ty]
                for k in o:
                    o[k] = decoder_hook_def(o[k])
                return cls(**o)
        return o


    global _Encoder, _decoder_hook
    _Encoder = EncoderDef
    _decoder_hook = decoder_hook_def


def test():
    from .fs import ImReadSetting
    from .dataset_reference import DatasetReference
    set = ImReadSetting()
    set_json = json.dumps(set, cls=get_encoder())
    set2: ImReadSetting = json.loads(set_json, object_hook=get_decoder_hook())
    assert set2.im_format == set.im_format
    dr = DatasetReference.empty()
    dr.im_read_setting = set
    dr_json = json.dumps(dr, cls=get_encoder())
    dr2: DatasetReference = json.loads(dr_json, object_hook=get_decoder_hook())
    assert dr.name == dr2.name



"""
This file handles file persistence of common objects used in pipeline. For example, dataset references
that specifies the files in a dataset can be stored as a file and read using the interfaces in this
file. Note that many file format is a directory containing a single file, which may seem that the directory
is unnecessary but this is to accommodate for the possible need to add more files to store information
about the object class' instance.
"""
import copy
import shutil
import json
from cvpl_tools.fs import ensure_dir_exists
from cvpl_tools.strenc import get_encoder, get_decoder_hook
from cvpl_tools.dataset_reference import DatasetReference


def write_dataset_reference(ref: DatasetReference, path: str):
    ensure_dir_exists(path, True)
    with open(f'{path}/dataset.json', 'w') as outfile:
        json.dump(ref, outfile, cls=get_encoder(), indent=2)


def read_dataset_reference(path: str) -> DatasetReference:
    with open(f'{path}/dataset.json', 'r') as infile:
        ref = json.load(infile, object_hook=get_decoder_hook())
    return ref


def write_dict(d, path: str, large_files=None):
    """
    Write a dictionary onto storage
    Commonly, a config/setting file will have path references to large objects. These objects
    can be copied to a nearby location to the config/setting file when the file is written, so
    this function provides a list to do such copying in addition to writting the dictionary:

    For each key, new_path in large_files, The object's corresponding path in dictionary, d[key]
    will be modified to new_path before the dictionary itself is written, and then the object
    will be copied to new_path. Then the dictionary will be written after all objects are copied.
    Args:
        d: The dictionary to be serialized
        path: The path to which the file will be stored
        large_files: A dictionary of large files to be written onto new locations
    """

    if large_files is None:
        large_files = {}
    ensure_dir_exists(path, True)
    d = copy.copy(d)
    for k in large_files:
        cur_path = d[k]
        target_path = large_files[k]
        if cur_path is None:
            raise ValueError(f'ERROR: Large file intended to be saved is not found from d[{k}]')
        d[k] = target_path
        shutil.copy(cur_path, target_path)
    with open(f'{path}/model_config.json', 'w') as outfile:
        json.dump(d, outfile, cls=get_encoder(), indent=2)


def read_dict(path: str):
    """
    Read a dictionary written by write_dict(). This will not read any large objects associated with
    the dictionary.
    Args:
        path: Path to the saved dictionary
    Returns:
        recovered dictionary object
    """
    with open(f'{path}/model_config.json', 'r') as infile:
        d = json.load(infile, object_hook=get_decoder_hook())
    return d



from datetime import datetime
from cvpl_tools.array_key_dict import ArrayKeyDict
import cvpl_tools.fs as fs
import numpy as np
from dataclasses import dataclass


@dataclass
class DatapointReference:
    """
    Represents a pointer to location of data.
    The data is either an image or a chunk (data point) in the dataset.
    """
    data_ref: str | list

    def __init__(self, data_ref: str | list):
        assert isinstance(data_ref, (str, list)), 'ERROR: Datapoint reference should be either a string or a list!'
        self.data_ref = data_ref

    def __str__(self):
        return str(self.data_ref)

    def ref(self):
        return self.data_ref

    def has_multiple_images(self):
        return isinstance(self.data_ref, list)

    def read_as_np(self, read_setting: fs.ImReadSetting) -> np.ndarray:
        return fs.ImIO.read_single_image(read_setting, self.data_ref)[0]


@dataclass
class DatasetReference:
    """
    This class' instance represents a dataset, and whose attributes include information on what data is in the dataset
    and when and how this reference is created.
    """
    datapoint_refs: ArrayKeyDict[str, DatapointReference]
    im_read_setting: fs.ImReadSetting
    creation_date: datetime
    creation_info: str  # a description of how this dataset is created
    name: str  # name of the dataset reference

    @classmethod
    def new(self, datapoint_refs: ArrayKeyDict[str, DatapointReference],
                 dataset_name,
                 creation_info,
                 im_read_setting: fs.ImReadSetting):
        # create an empty dataset
        return DatasetReference(
            datapoint_refs=datapoint_refs,
            creation_date=datetime.now(),
            creation_info=creation_info,
            name=dataset_name,
            im_read_setting=im_read_setting
        )

    @classmethod
    def empty(cls):
        return DatasetReference.new(ArrayKeyDict(),
                                'Empty Dataset',
                                'This dataset is created as an empty dataset.',
                                None)


def test():
    in1 = 'C:/path/to/file1'
    r1 = DatapointReference(in1)
    in2 = ['C:/path/to/file2', 'C:/path/to/file3']
    r2 = DatapointReference(in2)
    assert str(r1) == 'C:/path/to/file1'
    assert len(r2.ref()) == 2 and r2.ref()[0] == 'C:/path/to/file2'



import os


class ArrayKeyDict(dict):
    """
    This class ensures a dictionary with a customizable ordering imposed over the keys.
    Intended use cases additional to a Python dictionary:
    - Matching up two or more dictionaries' items by keys
    - Dictionary subsetting and merging
    - Modifying the key of an item while maintaining the position of the item in the dict
    - Looping over the keys in specific orders (may be different from the order they are inserted)
    - Obtain the i-th key in the array in O(1) time, where i is the key's position starting from 0

    Example:
        d = ArrayKeyDict([(3, 'a'), (4, 'b')])
        print(d.key_at(0))  # outputs 3
        for k in d:
            print(d[k], end='')  # outputs 'ab'
        print()
        d2 = ArrayKeyDict([(3, 'c'), (5, 'e')])
        comm_keys = d.common_keylist_with(d2)  # [3]
        d3 = d.subset_dict(comm_keys)  # now becomes [(3, 'a')]
        for k in d3:
            print(d3[k], end='')  # outputs 'a'
    """
    def __init__(self, pairs=()):
        """
        Create a dictionary; if the pairs parameter is provided, create a dictionary with items following the
        orders in the provided pairs
        :param pairs: An iterable of key, value pairs that specifies the initial items in the dictionary
        """
        super().__init__()
        self._key_arr = []

        # ensure ordering
        for k, v in pairs:
            self[k] = v

    def __setitem__(self, key, value):
        """
        Ensure the newly inserted value is at the end of key array;
        If the key already exists, do not change its order in the array
        """
        flag = key in self
        super().__setitem__(key, value)
        if not flag:
            self._key_arr.append(key)

    def keys(self) -> list:
        """
        Returns:
            A Python list, copy of the key array
        """
        return list(self._key_arr)

    def values(self) -> list:
        """
        Returns:
            A Python list, copy of the values in the order of the keys
        """
        return [self[k] for k in self._key_arr]

    def key_at(self, idx):
        """
        Obtain the idx-th index of the key array; does not check for out-of-bound access
        Args:
            idx: The index of the key to retrieve, index starting at 0
        Returns:
            The key at the idx-th index.
        """
        return self._key_arr[idx]

    def __iter__(self):
        """
        Returns:
            An iterator over the keys of the dictionary
        """
        return iter(self._key_arr)

    def pop(self, key, default=None):
        key_in_arr = key in self._key_arr
        if not key_in_arr:
            if default is None:
                raise KeyError(f'ERROR: Key {key} not found in array when pop is called!')
            else:
                return default

        self._key_arr.remove(key)
        item = super().pop(key)
        return item

    def __delitem__(self, key):
        key_in_arr = key in self._key_arr
        if not key_in_arr:
            raise KeyError(f'ERROR: Key {key} not found in array when del is called!')

        self._key_arr.remove(key)
        super().__delitem__(key)

    def rename_key(self, old_key, new_key):
        """
        Keeping the position and corresponding item of a key, but replacing itself with a different key.
        Args:
            old_key: The old key to be replaced
            new_key: The new key to replace the old key with
        """
        if new_key == old_key:
            return
        assert new_key not in self, f'ERROR: Attempt to rename to an existing key {new_key}'
        super().__setitem__(new_key, super().pop(old_key))
        for i in range(len(self._key_arr)):
            if self._key_arr[i] == old_key:
                self._key_arr[i] = new_key
                break

    def reorder_keys(self, ordered_key_arr):
        """
        Reorder the keys in the given order
        Args:
            ordered_key_arr: A list giving an ordering of the keys; all elements must appear exactly once
                in the original key list
        """
        # no new keys & length equal
        assert len(self._key_arr) == len(ordered_key_arr), f'ERROR: lengths of key arrays are different during reorder!'\
                                                           f' old {len(self._key_arr)} and new {len(ordered_key_arr)}'
        # no repetition
        assert len(ordered_key_arr) == len(set(ordered_key_arr)), f'ERROR: Repetition found in input {ordered_key_arr}!'
        # all keys are in the original dict
        for key in ordered_key_arr:
            assert key in self, f'ERROR: New key array contains value not present in the original key array: {key}'

        # all checking passed, we can safely replace the array
        self._key_arr = ordered_key_arr

    def subset_dict(self, subset_key_arr):
        """
        Similar to reorder_keys(), but with a smaller key array
        Args:
            subset_key_arr: A list of keys that specifies what are kept in the new ArrayKeyDict object
        Returns:
            a new ArrayKeyDict object containing only the items in the subset_key_arr, in the same order
        """
        # no new keys & length smaller
        assert len(self._key_arr) >= len(subset_key_arr), f'ERROR: subset key arr has larger length {len(subset_key_arr)}'\
                                                            f' than the original {len(self._key_arr)}'
        # no repetition
        assert len(subset_key_arr) == len(set(subset_key_arr)), f'ERROR: Repetition found in input {subset_key_arr}!'
        # all keys are in the original dict
        for key in subset_key_arr:
            assert key in self, f'ERROR: New key array contains value not present in the original key array: {key}'
        pairs = ((key, self[key]) for key in subset_key_arr)
        return ArrayKeyDict(pairs)

    def clear(self):
        """
        Clear all keys and items in the dictionary
        """
        self._key_arr.clear()
        super().clear()

    def trim_key_prefix_and_suffix(self, trim_front=0, trim_end=0):
        """
        For each key (of type str), rename it to a new str with both ends trimmed
        Args:
            trim_front: Number of characters to trim in front
            trim_end: Number of characters to trim in back
        """
        for i in range(len(self)):
            old_key = self._key_arr[i]
            new_key = old_key[trim_front:len(old_key) - trim_end]
            self.rename_key(old_key, new_key)

    def trim_key_common_prefix_and_suffix(self):
        """
        Similar to trim_key_prefix_and_suffix, but trim the redundant parts. In many cases the beginning
        and the end of keys are repetitive strings e.g. 'person1' and 'person2' the only identifiable part
        is '1' and '2', so the prefix 'person' can be trimmed to keep the key short.
        Among all keys, find common prefix and suffix and trim them off. This function modifies self.
        """
        keys = [s for s in self._key_arr]
        rkeys = [s[::-1] for s in self._key_arr]
        cprefix = os.path.commonprefix(keys)
        csuffix = os.path.commonprefix(rkeys)
        self.trim_key_prefix_and_suffix(len(cprefix), len(csuffix))

    def common_keylist_with(self, dicts):
        """
        Args:
            dicts: either an ArrayKeyDict or a list of ArrayKeyDict
        Returns:
            a Python list containing the common keys of this dictionary (self) and the provided other dictionary.
            The order of which the keys will appear in the result is the same as they are in self.
        """
        if type(dicts) is ArrayKeyDict:
            dicts = [dicts]
        keylist = []
        for key in self:
            is_common = True
            for d in dicts:
                if key not in d:
                    is_common = False
                    break
            if is_common:
                keylist.append(key)
        return keylist


def test():
    test_dict1 = ArrayKeyDict()
    test_dict2 = ArrayKeyDict([['file1', 1], ['file2', 3], ['file3', 2]])
    assert len(test_dict1) == 0 and len(test_dict2) == 3
    test_dict2['file2'] = 4
    assert len(test_dict2) == 3 and test_dict2['file2'] == 4
    test_dict2['file4'] = 2
    assert test_dict2.keys() == ['file1', 'file2', 'file3', 'file4'] and len(test_dict2) == 4
    assert test_dict2.key_at(2) == 'file3'

    for key in reversed(test_dict2):
        test_dict1[key] = test_dict2[key]
    assert test_dict1.keys() == ['file4', 'file3', 'file2', 'file1']
    test_dict1.rename_key('file1', 'f1')
    assert test_dict1.key_at(3) == 'f1' and test_dict1['f1'] == 1
    test_dict1.reorder_keys(['f1', 'file4', 'file2', 'file3'])
    assert test_dict1.key_at(1) == 'file4' and test_dict1.keys() == ['f1', 'file4', 'file2', 'file3']

    # prevent regression where rename_key() incorrectly push the renamed key to end of array
    test_dict1.rename_key('file2', 'f2')
    assert len(test_dict1) == 4  # prevent regression when rename_key adds length to _key_arr
    assert test_dict1.key_at(2) == 'f2'

    test_dict3 = ArrayKeyDict([['file1', 1], ['file2', 3], ['file3', 2]])
    assert test_dict1.common_keylist_with([test_dict2, test_dict3]) == ['file3']
    assert test_dict3.subset_dict(['file3', 'file1']).keys() == ['file3', 'file1']

    assert test_dict1.keys() == ['f1', 'file4', 'f2', 'file3']
    test_dict1.trim_key_common_prefix_and_suffix()
    assert test_dict1.key_at(0) == '1' and test_dict1.key_at(1) == 'ile4'

    test_dict1.clear()
    assert len(test_dict1) == 0 and len(test_dict2) == 4
    test_dict2.clear()
    assert len(test_dict1) == 0 and len(test_dict2) == 0

    print('All tests successfully completed')


import os
import shutil
import glob
import numpy as np
import nibabel as nib
from PIL import Image
import scipy
from dataclasses import dataclass, field
from cvpl_tools.array_key_dict import ArrayKeyDict

"""
fs.py
This is a file that provides methods to query and modify files in the file system
"""


def ensure_dir_exists(dir_path, remove_if_already_exists):
    if os.path.exists(dir_path):
        if remove_if_already_exists:
            shutil.rmtree(dir_path)
            os.mkdir(dir_path)
    else:
        os.mkdir(dir_path)


class ImFileType:
    FTYPE_UNKNOWN = -1
    FTYPE_JPX = 0
    FTYPE_JPG = 1
    FTYPE_PNG = 2
    FTYPE_NIB = 3
    FTYPE_TIF = 4
    FTYPE_NPY = 5
    FTYPE_NPZ = 6  # probabilities output by nnunet
    ext_to_ftype = {
        '.jpx': FTYPE_JPX,
        '.jpg': FTYPE_JPG,
        '.png': FTYPE_PNG,
        '.nii.gz': FTYPE_NIB,
        '.tif': FTYPE_TIF,
        '.npy': FTYPE_NPY,
        '.npz': FTYPE_NPZ,
    }
    ftype_to_ext = list(ext_to_ftype.keys())
    all_ftypes = list(ext_to_ftype.values())  # except unknown

    FORMAT_UINT8 = 0
    FORMAT_UINT16 = 1
    FORMAT_INT32 = 2
    FORMAT_FLOAT32 = 3
    format_to_np_dtype = [np.uint8, np.uint16, np.int32, np.float32]

    @classmethod
    def ftype_from_im_path(cls, im_path: str) -> int:
        """
        Infer the ftype from image path;
        The path can contain wildcards like "*"
        When the ftype can't be inferred, FTYPE_UNKNOWN will be returned
        """
        prefix, ext = os.path.splitext(im_path)
        ext = ext.lower()
        if ext == '.gz':
            prefix_prefix, prefix_ext = os.path.splitext(prefix)
            if prefix_ext.lower() == '.nii':
                return ImFileType.FTYPE_NIB
            else:
                return ImFileType.FTYPE_UNKNOWN
        elif ext in ImFileType.ext_to_ftype:
            return ImFileType.ext_to_ftype[ext]
        else:
            return ImFileType.FTYPE_UNKNOWN

    @classmethod
    def get_imid_and_zstack(cls, im_path: str, is_stack=False, keep_dirpath=False) -> tuple[str, str]:
        """
        Get the path stem as the unique image id from the path of the file;
        Second return value is the zstack number as a string
        file extensions will be stripped;
        If the file is part of a stacked image, then the z-stack numbers will be stripped from stem as will
        """
        ftype = ImFileType.ftype_from_im_path(im_path)
        if ftype == ImFileType.FTYPE_UNKNOWN:
            raise ValueError(f'Attempt to get the imid of an image file from path "{im_path}" with unknown file type!')
        elif ftype == ImFileType.FTYPE_NIB:
            imid = im_path[:-len('.nii.gz')]
        else:
            imid = os.path.splitext(im_path)[0]

        if not keep_dirpath:
            imid = os.path.split(imid)[1]

        if is_stack:
            i = -1
            while len(imid) >= -i and imid[i].isdigit():
                i -= 1
            return imid[:i + 1], imid[i + 1:]
        else:
            return imid, ''

    @classmethod
    def get_imid(cls, im_path: str, is_stack=False, keep_dirpath=False) -> str:
        return ImFileType.get_imid_and_zstack(im_path, is_stack, keep_dirpath)[0]

    @classmethod
    def load_im_as_np(cls, im_format, path, true_im_ndim: int, stack_axis=None,
                      concat_instead=False, allow_pickle=False) -> tuple:
        """
        Load individual file as a np array, or a stack of images as a np array if a list of path is given
        Returns a tuple (im, im_ftype, im_meta) where im_meta may be needed in order to save the im back to a file
        for certain file types
        """
        if isinstance(path, str):
            ftype = ImFileType.ftype_from_im_path(path)
            assert stack_axis is None, 'ERROR: Passing stack_axis in when loading single file'
            assert ftype != ImFileType.FTYPE_UNKNOWN, 'ERROR: Attempt to load an unknown image type!'
            if ftype == ImFileType.FTYPE_NPZ:
                assert im_format == ImFileType.FORMAT_FLOAT32
                im = np.load(path, allow_pickle=allow_pickle)['probabilities']
                im = np.transpose(im, (0, 3, 2, 1))
                im_meta = [ftype, im.shape, None, path, None]
            elif ftype == ImFileType.FTYPE_NPY:
                im = np.load(path, allow_pickle=allow_pickle)
                if allow_pickle:
                    im = im.item()['masks']
                im_meta = [ftype, im.shape, None, path, None]
            elif ftype == ImFileType.FTYPE_NIB:
                nii_im = nib.load(path)
                im = np.array(nii_im.get_fdata(), dtype=ImFileType.format_to_np_dtype[im_format])
                im_meta = [ftype, im.shape, None, path, nii_im.affine]
            else:
                im = np.array(Image.open(path), dtype=ImFileType.format_to_np_dtype[im_format])
                im_meta = [ftype, im.shape, None, path, None]

            assert im.dtype == ImFileType.format_to_np_dtype[im_format], \
                f'ERROR: Expect file type {ImFileType.format_to_np_dtype[im_format]} but got an image of dtype {im.dtype} at {path}'
            if len(im.shape) > true_im_ndim:
                axes = range(true_im_ndim, len(im.shape))
                for i in axes:
                    assert im.shape[i] == 1, f'Extra axes to true_im_ndim must be of width 1. true_im_ndim={true_im_ndim}, im.shape={im.shape}'
                im = im.squeeze(axis=tuple(axes))
            assert len(im.shape) == true_im_ndim, (f'ERROR: Expected true dimension {true_im_ndim}, but found '
                                                  f'image of shape {im.shape}!')
            assert type(im_meta) is list, 'ERROR: im_meta is not properly initialized!'
            return im, im_meta
        else:
            assert isinstance(path, list)
            im_arr = []
            im_meta = None
            for filepath in path:
                if concat_instead:
                    true_slice_dim = true_im_ndim
                else:
                    true_slice_dim = true_im_ndim - 1
                tup = ImFileType.load_im_as_np(im_format, filepath, true_slice_dim)
                im_arr.append(tup[0])
                cur_im_meta = tup[1]
                if im_meta is None:
                    im_meta = list(cur_im_meta)  # shallow copy
                    ftype_meta = im_meta[-1]
                    im_meta[-1] = [ftype_meta]
                else:
                    assert im_meta[0] == cur_im_meta[0]  # same file type
                    assert im_meta[1] == cur_im_meta[1]  # each image has the same shape
                    im_meta[-1].append(cur_im_meta[-1])
            if stack_axis is None:
                raise ValueError(f"Please specify stack_axis!")
                # stack_axis = true_im_ndim - 1  # default stack on the last axis
            if concat_instead:  # concat on existing axis
                assert 0 <= stack_axis < true_im_ndim - 1, f'Attempt to concat on axis {stack_axis} with true dim={true_im_ndim}'
                im = np.concatenate(im_arr, axis=stack_axis)
            else:  # stack on new axis
                assert 0 <= stack_axis < true_im_ndim, f'Attempt to concat on axis {stack_axis} with true dim={true_im_ndim}'
                im = np.stack(im_arr, axis=stack_axis)

            im_meta[2] = stack_axis
            im_meta[3] = path

            return im, im_meta

    @classmethod
    def save_np_as_im(cls, target_path: str, im: np.ndarray, ftype=None, write_axis_order=None, stack_axis=None,
                      concat_instead=False, ftype_meta=None, allow_pickle=False, im_format=None):
        """
        If ftype is None (default), it will be inferred from path
        """
        true_im_ndim = len(im.shape)
        if type(target_path) is str:
            assert not os.path.exists(target_path), f'{target_path} exists!'
            if ftype is None:
                ftype = ImFileType.ftype_from_im_path(target_path)
            if im_format is not None:
                assert im.dtype == ImFileType.format_to_np_dtype[im_format]
            assert ftype != ImFileType.FTYPE_UNKNOWN, 'ERROR: Attempt to save an unknown image type!'
            if ftype == ImFileType.FTYPE_NPZ:
                assert im.dtype == np.float32
                im = np.transpose(im, (0, 3, 2, 1))
                np.savez(target_path, probabilities=im, allow_pickle=allow_pickle)
            elif ftype == ImFileType.FTYPE_NPY:
                if allow_pickle:
                    im = {'masks': im}
                np.save(target_path, im, allow_pickle=allow_pickle)
            elif ftype == ImFileType.FTYPE_NIB:
                assert ftype_meta.shape == (4, 4), f'Affine should have shape 4, 4, but get {ftype_meta[-1].shape}'
                nii_im = nib.Nifti1Image(im, ftype_meta)
                nib.save(nii_im, target_path)
            else:
                Image.fromarray(im).save(target_path)
        else:
            if stack_axis is None:
                stack_axis = true_im_ndim - 1
            for i in range(im.shape[stack_axis]):
                if concat_instead:
                    im_slice = im.take(indices=range(i, i + 1), axis=stack_axis)
                else:
                    im_slice = im.take(indices=i, axis=stack_axis)
                target = target_path[i]
                if ftype_meta is None:
                    cur_ftype_meta = None
                else:
                    cur_ftype_meta = ftype_meta[i]
                # recursive call!
                ImFileType.save_np_as_im(
                    target_path=target,
                    im=im_slice,
                    ftype=ftype,
                    write_axis_order=write_axis_order,
                    stack_axis=stack_axis,
                    concat_instead=False,
                    ftype_meta=cur_ftype_meta,
                    allow_pickle=allow_pickle)


@dataclass
class ImReadSetting:
    """
    im_format (int) - specifies the data storage layout e.g. ImFileType.FORMAT_UINT8
    stack_axis (int) - if not None, combine several files with different zstacks to the same imid; specify the axis to stack on
    concat_instead (bool) - if True, the stacking are done by np.concatenate() which preserves im dimensions; over the same axis
    true_im_ndim (int) - # of dimensions of a single image; includes number of channels (for 2d BGR image this would be 3)
    allow_pickle (bool) - set allow_pickle for np.load and np.save
    imid_keep_dirpath (bool) - keeping the directory path when extracting imid from paths; default to False
    """
    im_format: int = ImFileType.FORMAT_UINT8
    stack_axis: int = None
    concat_instead: bool = False
    true_im_ndim: int = 3
    allow_pickle: bool = False
    imid_keep_dirpath: bool = False


@dataclass
class ImWriteSetting:
    """
    im_format (int) - specifies the data storage layout e.g. ImFileType.FORMAT_UINT8
    stack_axis (int) - if not None, combine several files with different zstacks to the same imid; specify the axis to stack on
    concat_instead (bool) - if True, the stacking are done by np.concatenate() which preserves im dimensions; over the same axis
    allow_pickle (bool) - set allow_pickle for np.load and np.save
    ftype (int) - file type to write e.g. ImFileType.FTYPE_NPZ
    """
    im_format: int = ImFileType.FORMAT_UINT8
    stack_axis: int = None
    concat_instead: bool = False
    allow_pickle: bool = False
    ftype: int = None


class ImIO:
    @classmethod
    def read_single_image(cls, read_setting: ImReadSetting, path):
        im, im_meta = ImFileType.load_im_as_np(read_setting.im_format, path,
                                               true_im_ndim=read_setting.true_im_ndim,
                                               stack_axis=read_setting.stack_axis,
                                               concat_instead=read_setting.concat_instead,
                                               allow_pickle=read_setting.allow_pickle)
        return im, im_meta

    @classmethod
    def write_single_image(cls, write_setting: ImWriteSetting, path, im, im_meta):
        ImFileType.save_np_as_im(path, im=im, ftype=write_setting.ftype,
                                 write_axis_order=write_setting.write_axis_order,
                                 stack_axis=write_setting.stack_axis,
                                 concat_instead=write_setting.concat_instead,
                                 ftype_meta=im_meta[-1],
                                 allow_pickle=write_setting.allow_pickle)

    @classmethod
    def read_filenames(cls, read_setting: ImReadSetting, pattern: str = None, path=None):
        """
        The keys (imids) of the ArrayKeyDict returned by this function will be sorted alphabetically
        Parameters
            path (list) - optional and passed in as a list of all paths if pattern is None
        Return
            imid_to_path (Dict) - maps from imids to either a filename or a list of filenames (if image is stacked);
                The keys of this dictionary is a sorted list of image imids, act as unique ids of images, sorted alphabetically
        """
        imid_to_path = ImIO._get_imid_to_path_mapping(read_setting, path, pattern)
        imid_to_path.reorder_keys(sorted(imid_to_path.keys()))
        return imid_to_path

    @classmethod
    def read_image(cls, read_setting: ImReadSetting, paths: list):
        """
        Parameter
            paths (list) - can be nested in case of stacked images
        Return
            an iterator whose elements are (im, im_meta), im is the np array of the image and im_meta contains much
            additional metadata returned by ImFileType.load_im_as_np()
        """
        assert type(paths) is not ArrayKeyDict, f'Attempt to pass an ArrayKeyDict to read_image()!'
        for path in paths:
            im, im_meta = ImIO.read_single_image(read_setting, path)
            yield im, im_meta

    @classmethod
    def write_image(cls, write_setting: ImWriteSetting, paths: list, im_iter: iter):
        """
        Parameter
            paths (list) - can be nested in case of stacked images
            im_iter (iter) - An iterator returning tuples (im, im_meta), the order of these images should be the same as
                the order of their paths in the paths variable
        """
        for path in paths:
            im, im_meta = next(im_iter)
            ImIO.write_single_image(write_setting, path, im, im_meta)

    @classmethod
    def _get_imid_to_path_mapping(cls, read_setting: ImReadSetting,
                                  paths: list[str] = None, pattern: str = None):
        if paths is None:
            paths = glob.glob(pattern)
        elif pattern is not None:
            print('WARNING: both paths and pattern are not None! Is this intended?')

        imid_to_path = ArrayKeyDict()
        for path in paths:
            ftype = ImFileType.ftype_from_im_path(path)
            if ftype == ImFileType.FTYPE_UNKNOWN:
                continue
            imid, zstack_str = ImFileType.get_imid_and_zstack(path, is_stack=read_setting.stack_axis is not None,
                                                              keep_dirpath=read_setting.imid_keep_dirpath)
            if zstack_str == '':
                imid_to_path[imid] = path
            else:
                tup = (zstack_str, path)
                if imid not in imid_to_path:
                    imid_to_path[imid] = [tup]
                else:
                    imid_to_path[imid].append(tup)
        for k in list(imid_to_path.keys()):
            v = imid_to_path[k]
            if type(v) is list:
                imid_to_path[k] = [tup[1] for tup in sorted(v)]
        return imid_to_path


def im_resize(target_size: tuple, im: np.ndarray, order: int = 0) -> np.ndarray:
    assert len(im.shape) == len(target_size), f'Image has shape {im.shape} but intend to resize ' \
                                              f'to incompatible shape {target_size}'
    zoom_factors = (target_size[i] / im.shape[i] for i in range(len(target_size)))
    resized_im = scipy.ndimage.zoom(im, zoom_factors, order=order)  # order=0 -> use nearest neighbor
    return resized_im


def test():
    assert ImFileType.ftype_from_im_path('./some_root/hello.nii.gz') == ImFileType.FTYPE_NIB
    assert ImFileType.ftype_from_im_path('./some_root/hello_0001.JPG') == ImFileType.FTYPE_JPG
    assert ImFileType.ftype_from_im_path('./a.py') == ImFileType.FTYPE_UNKNOWN
    assert ImFileType.get_imid('./some_root/hello.png') == 'hello'
    assert (ImFileType.get_imid('./some_root/nibfile0001.nii.gz', is_stack=True, keep_dirpath=True) ==
            './some_root/nibfile')

    print('All tests successfully completed')



from cvpl_tools.fs import ImFileType, ImReadSetting, ImIO


def reader_from_cmd(cmd: list, true_im_dim, uint8_only=False):
    fm = ImFileType.FORMAT_UINT8
    stack_axis = None
    concat_instead = False
    path_pattern = ''
    trim_front = 0
    trim_end = 0
    trim_common = False
    for i in range(0, len(cmd)):
        term = cmd[i]
        if i == 0:
            path_pattern = term
        elif term == '--probs':
            if uint8_only:
                raise ValueError('ERROR: Attempt to specify "--probs" option (implicit float32) when loading a uint8 '
                                 'image!')
            fm = ImFileType.FORMAT_FLOAT32
        elif term == '--stack':
            stack_axis = int(cmd[i + 1])
        elif term == '--concat':
            stack_axis = int(cmd[i + 1])
            concat_instead = True
        elif term == '--ord':
            if uint8_only:
                raise ValueError('ERROR: Attempt to specify "--ord" option (implicit uint16) when loading a uint8 '
                                 'image!')
            fm = ImFileType.FORMAT_UINT16
        elif term == '--trim_front':
            trim_front = int(cmd[i + 1])
        elif term == '--trim_end':
            trim_end = int(cmd[i + 1])
        elif term == '--trim_common':
            trim_common = True
    im_read_setting = ImReadSetting(
        im_format=fm,
        true_im_dim=true_im_dim,
        stack_axis=stack_axis,
        concat_instead=concat_instead,
        allow_pickle=True)
    imid_to_paths = ImIO.read_filenames(im_read_setting, path_pattern)
    if trim_front != 0 or trim_end != 0:
        imid_to_paths.trim_key_prefix_and_suffix(trim_front, trim_end)
    if trim_common:
        imid_to_paths.trim_key_common_prefix_and_suffix()

    return path_pattern, im_read_setting, imid_to_paths